"""Utilities for handling the XML output of gccxml __FOR A C
LIBRARY__. If you need a C++ parser, you're better of with pygccxml,
which generates Boost.Python classes.

Requires python 2.5
"""

##############################################################################

import xml.etree.cElementTree as ET
import ctypes
import sys

def load_file(filename):
    return ET.parse(filename)

class CModule(object):
    pass

class Unsupported(Exception):
    pass

class Header(object):
    def __init__(self,
                 file_name,
                 library):
        self.tree = ET.parse(file_name)
        self.library = library
        self._elements = {}
        self._verbose = False
        for child in self.tree.getroot().getchildren():
            try:
                self._elements[child.attrib['id']] = child
            except AttributeError:
                pass
        self.types = Types(self)
    
    def files(self):
        return self.tree.getroot().findall('./File')

    def create_module(self, files_to_use):
        new_module = CModule()

        # create list of files to use
        ok_files = set()
        for child in self.files():
            if not any(child.attrib['name'].find(f) != -1
                       for f in files_to_use):
                continue
            ok_files.add(child.attrib['id'])
        
        # variables
        for child in self.tree.getroot().findall('./Variable'):
            if child.attrib['file'] not in ok_files:
                # defined somewhere else, skip
                continue
            try:
                t = self.types.resolve(child.attrib['type'])
            except Unsupported, e:
                if not self._verbose:
                    continue
                print >>sys.stderr, ("WARNING: skipping %s because %s" %
                                     (child.attrib['name'],
                                      e.message))
                continue
            try:
                var = t.in_dll(self.library, child.attrib['name'])
            except ValueError:
                if not self._verbose:
                    continue
                print >>sys.stderr, ("WARNING: Could not find symbol '%s'. Skipping." %
                                     child.attrib['name'])
            else:
                setattr(new_module, child.attrib['name'], var)
        
        # enumerations
        for child in self.tree.getroot().findall('./Enumeration'):
            if child.attrib['file'] not in ok_files:
                # defined somewhere else, skip
                continue
            for gc in child.findall('./EnumValue'):
                setattr(new_module, gc.attrib['name'], int(gc.attrib['init']))
        
        # functions
        for child in self.tree.getroot().findall('./Function'):
            if child.attrib['file'] not in ok_files:
                continue
            try:
                f = getattr(self.library, child.attrib['name'])
            except AttributeError:
                if not self._verbose:
                    continue
                print >>sys.stderr, ("WARNING: Could not find symbol '%s'. Skipping." %
                                     child.attrib['name'])
                continue
            f.restype = self.types.resolve(child.attrib['returns'])
            # Only set argtypes if no varargs
            if len(child.findall('Ellipsis')) == 0:
                argtypes = []
                for gc in child.findall('Argument'):
                    argtypes.append(self.types.resolve(gc.attrib['type']))
                f.argtypes = argtypes
            setattr(new_module, child.attrib['name'], f)

        # structure types
        for child in self.tree.getroot().findall('./Struct'):
            if child.attrib['file'] not in ok_files:
                continue
            t = self.types.resolve(child.attrib['id'])
            setattr(new_module, child.attrib['name'], t)
            # Also set various indirection level pointers to the structures
            setattr(new_module, child.attrib['name'] + '_p', ctypes.POINTER(t))
            setattr(new_module, child.attrib['name'] + '_pp', ctypes.POINTER(ctypes.POINTER(t)))
            setattr(new_module, child.attrib['name'] + '_ppp', ctypes.POINTER(ctypes.POINTER(ctypes.POINTER(t))))

        # typedefs
        for child in self.tree.getroot().findall('./Typedef'):
            if child.attrib['file'] not in ok_files:
                continue
            t = self.types.resolve(child.attrib['id'])
            setattr(new_module, child.attrib['name'], t)

        return new_module

class Types(object):

    def __init__(self, header):
        self._header = header
        self._types = {}

    def resolve_ReferenceType(self, element):
        raise Unsupported("Reference types not supported")

    def resolve_FunctionType(self, element):
        params = [self.resolve(element.attrib['returns'])]
        if len(element.findall('./Ellipsis')):
            raise Unsupported("ctypes can't deal with function pointers to varargs")
        for arg in element.findall('./Argument'):
            params.append(self.resolve(arg.attrib['type']))
        return ctypes.CFUNCTYPE(*params)
            
    def resolve_Typedef(self, element):
        return self.resolve(element.attrib['type'])

    # ctypes ignores const, let's treat it as a typedef
    resolve_CvQualifiedType = resolve_Typedef

    def resolve_ArrayType(self, element):
        try:
            array_size = int(element.attrib['max']) - int(element.attrib['min']) + 1
        except ValueError:
            return ctypes.POINTER(self.resolve(element.attrib['type']))
        return self.resolve(element.attrib['type']) * array_size

    def resolve_PointerType(self, element):
        return ctypes.POINTER(self.resolve(element.attrib['type']))

    def resolve_FundamentalType(self, element):
        if element.attrib['id'] in self._types:
            return self._types[element.attrib['id']]
        d = {'double': ctypes.c_double,
             'float': ctypes.c_float,
             'int': ctypes.c_int,
             'unsigned int': ctypes.c_uint,
             'long long unsigned int': ctypes.c_ulonglong,
             'long long int': ctypes.c_longlong,
             'long unsigned int': ctypes.c_ulong,
             'long int': ctypes.c_long,
             'short unsigned int': ctypes.c_ushort,
             'short int': ctypes.c_short,
             'unsigned char': ctypes.c_byte,
             'signed char': ctypes.c_char,
             'char': ctypes.c_char,
             'void': None,
             'long double': ctypes.c_double * 2, # THIS IS WRONG
             'wchar_t': ctypes.c_wchar,
             'complex long double': ctypes.c_double * 4, # THIS IS WRONG
             'complex double': ctypes.c_double * 2,
             'complex float': ctypes.c_float * 2}
        return d[element.attrib['name']]

    def resolve_Union(self, element):
        new_type = type(element.attrib['name'],
                        (ctypes.Union,),{})
        # because unions can have pointers to themselves, we need this hack.

        try:
            member_ids = element.attrib['members'].split()
        except KeyError:
            # this is an opaque struct
            return new_type

        new_type._fields_ = []
        for member_id in member_ids:
            member = self._header._elements[member_id]
            # This is C, we don't care about constructors
            if member.tag != 'Field':
                continue
            new_type._fields_.append((member.attrib['name'],
                                      self.resolve(member.attrib['type'])))
        return new_type

    def resolve_Struct(self, element):
        new_type = type(element.attrib['name'],
                        (ctypes.Structure,),{})
        # because structs can have pointers to themselves, we need this hack.
        self._types[element.attrib['id']] = new_type

        try:
            member_ids = element.attrib['members'].split()
        except KeyError:
            # this is an opaque struct
            return new_type

        _fields_ = []
        for member_id in member_ids:
            member = self._header._elements[member_id]
            # This is C, we don't care about constructors
            if member.tag != 'Field':
                continue
            _fields_.append((member.attrib['name'],
                             self.resolve(member.attrib['type'])))
        new_type._fields_ = _fields_
        return new_type

    def resolve(self, type_id):
        if type_id in self._types:
            return self._types[type_id]
        el = self._header._elements[type_id]
        result = getattr(self, 'resolve_' + el.tag)(el)
        self._types[type_id] = result
        return result
