##############################################################################
# gage class module
# 
# Part of the python bindings for teem, by Carlos Scheidegger
# teem is developed by Gordon Kindlmann, and available at
# http://teem.sourceforge.net

"""Python classes for gage, part of the teem library.

"""

from base import *
import teem.capi as c
import teem.nrrd as nrrd
import teem
import copy
import itertools

##############################################################################

class Kind(object):
    Scalar = c.gageKindScl
    Vector = c.gageKindVec
    
gageKindScl = c.gageKindScl
gageKindVec = c.gageKindVec
GAGE_ITEM_MAX = c.GAGE_ITEM_MAX
GAGE_PERVOLUME_MAXNUM = c.GAGE_PERVOLUME_MAXNUM

class Kernel(object):
    Unknown = c.gageKernelUnknown
    K00 = c.gageKernel00
    K10 = c.gageKernel10
    K11 = c.gageKernel11
    K20 = c.gageKernel20
    K21 = c.gageKernel21
    K22 = c.gageKernel22
    Stack = c.gageKernelStack

class ScalarMeasurement(object):
    Unknown = c.gageSclUnknown      #   0: nobody knows 
    Value = c.gageSclValue        #   1: "v", data value: [1] 
    GradVec = c.gageSclGradVec      #   2: "grad", gradient vector, un-normalized: [3] 
    GradMag = c.gageSclGradMag      #   3: "gm", gradient magnitude: [1] 
    Normal = c.gageSclNormal       #   4: "n", gradient vector, normalized: [3] 
    NPerp = c.gageSclNPerp        #   5: "np", projection onto tangent plane: [9] 
    Hessian = c.gageSclHessian      #   6: "h", Hessian: [9] (column-order) 
    Laplacian = c.gageSclLaplacian    #   7: "l", Laplacian: Dxx + Dyy + Dzz: [1] 
    HessFrob = c.gageSclHessFrob     #   8: "hf", Frobenius normal of Hessian: [1] 
    HessEval = c.gageSclHessEval     #   9: "heval", Hessian's eigenvalues: [3] 
    HessEval0 = c.gageSclHessEval0    #  10: "heval0", Hessian's 1st eigenvalue: [1] 
    HessEval1 = c.gageSclHessEval1    #  11: "heval1", Hessian's 2nd eigenvalue: [1] 
    HessEval2 = c.gageSclHessEval2    #  12: "heval2", Hessian's 3rd eigenvalue: [1] 
    HessEvec = c.gageSclHessEvec     #  13: "hevec", Hessian's eigenvectors: [9] 
    HessEvec0 = c.gageSclHessEvec0    #  14: "hevec0", Hessian's 1st eigenvector: [3] 
    HessEvec1 = c.gageSclHessEvec1    #  15: "hevec1", Hessian's 2nd eigenvector: [3] 
    HessEvec2 = c.gageSclHessEvec2    #  16: "hevec2", Hessian's 3rd eigenvector: [3] 
    SndDD = c.gageScl2ndDD        #  17: "2d", 2nd dir.deriv. along gradient: [1] 
    GeomTens = c.gageSclGeomTens     #  18: "gten", sym. matx w/ evals {0, K1, K2} and
    #       evecs {grad, cdir0, cdir1}: [9]
    K1 = c.gageSclK1           #  19: "k1", 1st principle curvature: [1] 
    K2 = c.gageSclK2           #  20: "k2", 2nd principle curvature (k2 <= k1): [1] 
    TotalCurv = c.gageSclTotalCurv    #  21: "tc", L2 norm(K1,K2) (not Koen.'s "C"): [1] 
    ShapeTrace = c.gageSclShapeTrace   #  22, "st", (K1+K2)/Curvedness: [1] 
    ShapeIndex = c.gageSclShapeIndex   #  23: "si", Koen.'s shape index, ("S"): [1] 
    MeanCurv = c.gageSclMeanCurv     #  24: "mc", mean curvature (K1 + K2)/2: [1] 
    GaussCurv = c.gageSclGaussCurv    #  25: "gc", gaussian curvature K1*K2: [1] 
    CurvDir1 = c.gageSclCurvDir1     #  26: "cdir1", 1st principle curv direction: [3] 
    CurvDir2 = c.gageSclCurvDir2     #  27: "cdir2", 2nd principle curv direction: [3] 
    FlowlineCurv = c.gageSclFlowlineCurv #  28: "fc", curvature of normal streamline: [1] 
    Median = c.gageSclMedian       #  29: "med", median filter 
    HessValleyness = c.gageSclHessValleyness   #  30: "valley measure", vallyness measure: [1] 
    HessRidgeness = c.gageSclHessRidgeness    #  31: "ridge measure", ridgeness measure: [1] 
    HessMode = c.gageSclHessMode     #  32: "mode hessian", Hessian's mode: [1]     

class VectorMeasurement(object):
    Unknown = c.gageVecUnknown #        /*  0: nobody knows */
    Vector = c.gageVecVector #         /*  1: "v", component-wise-intrpolated
    #                                 (CWI) vec: [3] */
    Vector0 = c.gageVecVector0 #        /*  2: "v0", vector[0]: [1] */
    Vector1 = c.gageVecVector1 #        /*  3: "v1", vector[0]: [1] */
    Vector2 = c.gageVecVector2 #        /*  4: "v2", vector[0]: [1] */
    Length = c.gageVecLength #         /*  5: "l", length of CWI vector: [1] */
    Normalized = c.gageVecNormalized #     /*  6: "n", normalized CWI vector: [3] */
    Jacobian = c.gageVecJacobian #       /*  7: "j", component-wise Jacobian: [9]
                               # 0:dv_x/dx  1:dv_x/dy  2:dv_x/dz
                               # 3:dv_y/dx  4:dv_y/dy  5:dv_y/dz
                               # 6:dv_z/dx  7:dv_z/dy  8:dv_z/dz */
    Divergence = c.gageVecDivergence #     /*  8: "d", divergence (based on Jacobian): [1] */
    Curl = c.gageVecCurl #           /*  9: "c", curl (based on Jacobian): [3] */
    CurlNorm = c.gageVecCurlNorm #       /* 10: "cm", curl magnitude: [1] */
    Helicity = c.gageVecHelicity #       /* 11: "h", helicity: vec . curl: [1] */
    NormHelicity = c.gageVecNormHelicity #   /* 12: "nh", normalized helicity: [1] */
    Lambda2 = c.gageVecLambda2 #        /* 13: "lambda2", lambda2 criterion: [1] */
    ImaginaryPart = c.gageVecImaginaryPart #  /* 14: "imag", imag. part of jacobian's
                               # eigenv: [1] */
    Hessian = c.gageVecHessian #        /* 15: "vh", second-order derivative: [27] 
                               #  HEY: indices here need to be double checked
                               #  0:d2v_x/dxdx   1:d2v_x/dxdy   2:d2v_x/dxdz
                               #  3:d2v_x/dydx   4:d2v_x/dydy   5:d2v_x/dydz
                               #  6:d2v_x/dzdx   7:d2v_x/dzdy   8:d2v_x/dzdz
                               #  9:d2v_y/dxdx       [..]
                               #     [..]
                               # 24:dv2_z/dzdx  25:d2v_z/dzdy  26:d2v_z/dzdz */
    DivGradient = c.gageVecDivGradient #    /* 16: "dg", divergence gradient: [3] */
    CurlGradient = c.gageVecCurlGradient #   /* 17: "cg", curl gradient: [9] */
    CurlNormGrad = c.gageVecCurlNormGrad #   /* 18: "cng", curl norm gradient: [3] */
    NCurlNormGrad = c.gageVecNCurlNormGrad #  /* 19: "ncng", normalized curl norm gradient: [3] */
    HelGradient = c.gageVecHelGradient #    /* 20: "hg", helicity gradient: [3] */
    DirHelDeriv = c.gageVecDirHelDeriv #    /* 21: "dhd", directional derivative
                                 # of helicity: [1] */ 
    ProjHelGradient = c.gageVecProjHelGradient #/* 22: "phg", projected helicity gradient: [3] */
    Gradient0 = c.gageVecGradient0 #      /* 23: "g0", gradient of 1st coeff of vector: [3] */
    Gradient1 = c.gageVecGradient1 #      /* 24: "g1", gradient of 2nd coeff of vector: [3] */
    Gradient2 = c.gageVecGradient2 #      /* 25: "g2", gradient of 3rd coeff of vector: [3] */
    MultiGrad = c.gageVecMultiGrad #      /* 26: "mg", sum of outer products of grads: [9] */
    MGFrob = c.gageVecMGFrob #         /* 27: "mgfrob", frob norm of multi-gradient: [1] */
    MGEval = c.gageVecMGEval #         /* 28: "mgeval", evals of multi-gradient: [3] */
    MGEvec = c.gageVecMGEvec #         /* 29: "mgevec", evecs of multi-gradient: [9] */

##############################################################################

class GageContext(object):

    ##########################################################################
    
    def __init__(self):
        self._cgagecontext = c.gageContextNew()

    def __del__(self):
        c.gageContextNix(self._ctype)

    def _get_ctype(self):
        return self._cgagecontext
    _ctype = property(_get_ctype)

    ##########################################################################

    def attach(self, pvl):
        c.gagePerVolumeAttach(self._ctype, pvl._ctype)
        pvl._is_attached = True

    def set_kernel(self, which, kernel, params):
        p = (c_double * nrrd.NRRD_KERNEL_PARMS_NUM)()
        for i, v in enumerate(params):
            p[i] = v
        c.gageKernelSet(self._ctype, which, kernel, p)

    def query_item_on(self, pvl, gage_kind):
        c.gageQueryItemOn(self._ctype, pvl._ctype, gage_kind)

    def update(self):
        c.gageUpdate(self._ctype)

    def answer_pointer(self, pvl, gage_kind):
        return c.gageAnswerPointer(self._ctype, pvl._ctype, gage_kind)

    def probe(self, x, y, z):
        c.gageProbe(self._ctype, x, y, z)

    def probe_many(self, nin):
        nrrds = ((c.Nrrd_p * GAGE_ITEM_MAX) * GAGE_PERVOLUME_MAXNUM)()
        c.gageProbeMany(self._ctype, nrrds, nin._ctype)
        result = []
        for i in xrange(self._ctype.contents.pvlNum):
            m = {}
            for j in xrange(GAGE_ITEM_MAX):
                if cast(nrrds[i][j], c_void_p).value is not None:
                    m[j] = nrrd.NrrdC(nrrds[i][j], True)
            result.append(m)
        return result

class GagePerVolume(object):

    ##########################################################################

    def __init__(self, ctx, nin, kind):
        self._cgagepervolume = c.gagePerVolumeNew(ctx._ctype, nin._ctype, kind)
        self._is_attached = False

    def __del__(self):
        if not self._is_attached:
            c.gagePerVolumeNix(self._cgagepervolume)

    def _get_ctype(self):
        return self._cgagepervolume
    _ctype = property(_get_ctype)

    ##########################################################################
    
def check_volume(kind, nin):
    c.gageKindVolumeCheck(kind, nin._ctype)
