##############################################################################
# nrrd class module
# 
# Part of the python bindings for teem, by Carlos Scheidegger
# teem is developed by Gordon Kindlmann, and available at
# http://teem.sourceforge.net

"""Python classes for NRRD, part of the teem library.

This exposes most of the NRRD 'bulk' functionality. Individual member
access is not currently possible. It is easily added, but if you're
accessing things individually, it'll be slow enough that you probably
want to do it in C anyway.

Notably, this library converts between NRRDs and numpy arrays. This
allows one to use the excellent scipy/numpy/matplotlib packages.

The biggest gotcha here is that NRRDs go fastest axis first, and
numpy goes slowest axis first, so

>>> x
array([[ 1.,  2.,  3.,  4.],
       [ 5.,  6.,  7.,  8.]])
>>> numpy.array(teem.nrrd.Nrrd.from_array(x).sum(0))
array([ 10.,  26.])
>>> x.sum(0)
array([  6.,   8.,  10.,  12.])

"""

from base import *
import teem.capi as c
import teem.capi.numpy
import teem
import copy
import itertools

##############################################################################
# Defines

from teem.capi import (
    NRRD,
    NRRD_DIM_MAX,
    NRRD_SPACE_DIM_MAX,
    NRRD_KERNEL_PARMS_NUM)

##############################################################################
# Nrrd types

class Type(object):
    Default = c.nrrdTypeDefault
    Char =    c.nrrdTypeChar    
    UChar =   c.nrrdTypeUChar   
    Short =   c.nrrdTypeShort   
    UShort =  c.nrrdTypeUShort  
    Int =     c.nrrdTypeInt     
    UInt =    c.nrrdTypeUInt    
    LLong =   c.nrrdTypeLLong   
    ULLong =  c.nrrdTypeULLong  
    Float =   c.nrrdTypeFloat   
    Double =  c.nrrdTypeDouble

##############################################################################
# Nrrd measure types for projection

class Measure(object):
    Min           = c.nrrdMeasureMin            #  1: smallest value 
    Max           = c.nrrdMeasureMax            #  2: biggest value 
    Mean          = c.nrrdMeasureMean           #  3: average of values 
    Median        = c.nrrdMeasureMedian         #  4: value at 50th percentile 
    Mode          = c.nrrdMeasureMode           #  5: most common value 
    Product       = c.nrrdMeasureProduct        #  6: product of all values 
    Sum           = c.nrrdMeasureSum            #  7: sum of all values 
    L1            = c.nrrdMeasureL1             #  8 
    L2            = c.nrrdMeasureL2             #  9 
    Linf          = c.nrrdMeasureLinf           #  10 
    Variance      = c.nrrdMeasureVariance       #  11 
    SD            = c.nrrdMeasureSD             #  12: standard deviation 
    Skew          = c.nrrdMeasureSkew           #  13: skew 
    LineSlope     = c.nrrdMeasureLineSlope      #  14: slope of line of best fit 
    LineIntercept = c.nrrdMeasureLineIntercept  #  15: y-intercept of line of best fit 
    LineError     = c.nrrdMeasureLineError      #  16: error of line fitting 
    HistoMin      = c.nrrdMeasureHistoMin       #  17 
    HistoMax      = c.nrrdMeasureHistoMax       #  18 
    HistoMean     = c.nrrdMeasureHistoMean      #  19 
    HistoMedian   = c.nrrdMeasureHistoMedian    #  20 
    HistoMode     = c.nrrdMeasureHistoMode      #  21 
    HistoProduct  = c.nrrdMeasureHistoProduct   #  22 
    HistoSum      = c.nrrdMeasureHistoSum       #  23 
    HistoL2       = c.nrrdMeasureHistoL2        #  24 
    HistoVariance = c.nrrdMeasureHistoVariance  #  25 
    HistoSD       = c.nrrdMeasureHistoSD        #  26 

    
##############################################################################
# Nrrd boundary behaviors

class Boundary(object):
    Pad    = c.nrrdBoundaryPad
    Bleed  = c.nrrdBoundaryBleed
    Wrap   = c.nrrdBoundaryWrap
    Weight = c.nrrdBoundaryWeight

##############################################################################
# Nrrd formats and encodings (not all match all)

class Format(object):
    NRRD = c.nrrdFormatNRRD
    PNM  = c.nrrdFormatPNM
    PNG  = c.nrrdFormatPNG
    VTK  = c.nrrdFormatVTK
    Text = c.nrrdFormatText
    EPS  = c.nrrdFormatEPS

class Encoding(object):
    Raw   = c.nrrdEncodingRaw
    Ascii = c.nrrdEncodingAscii
    Hex   = c.nrrdEncodingHex
    Gzip  = c.nrrdEncodingGzip
    Bzip2 = c.nrrdEncodingBzip2

##############################################################################
# Utility

def nrrd_fun_make_new_nrrd(nrrd_fun):
    def do_it(*args):
        # make new nrrd
        r = Nrrd(True)
        new_args = []
        for arg in args:
            # if object exposes a _ctype accessor, use it
            try:
                new_args.append(arg._ctype)
            except AttributeError:
                new_args.append(arg)
        nrrd_fun(r._cnrrd, *new_args)
        return r
    return do_it

##############################################################################
# NrrdRange class

class NrrdRangeBase(object):

    def __del__(self):
        c.nrrdRangeNix(self._cnrrdrange)

    def _get_ctype(self):
        return self._cnrrdrange
    _ctype = property(_get_ctype)

    ##########################################################################

    min = ctype_simple_rw_property('min')
    max = ctype_simple_rw_property('max')

    def __copy__(self):
        # ugly
        new_ctype_obj = c.nrrdRangeCopy(self._cnrrdrange)
        return NrrdRangeC(new_ctype_obj)

    ##########################################################################

    def reset(self):
        c.nrrdRangeReset(self._cnrrdrange)

    def set(self, nrrd, blind8BitRange):
        c.nrrdRangeSet(self._cnrrdrange, nrrd._cnrrd, blind8bitRange)

    def safe_set(self, nrrd, blind8bitRange):
        c.nrrdRangeSafeSet(self._cnrrdrange, nrrd._cnrrd, blind8bitRange)

# NrrdRange is for things that are created on the Python side of the
# Python-C boundary
class NrrdRange(NrrdRangeBase):

    def __init__(self, *args):
        if type(args[0]) == float:
            # standard nrrdRangeNew construction
            min, max = args
            self._cnrrdrange = c.nrrdRangeNew(min, max)
        else:
            # nrrd construction
            try:
                nrrd, blind8BitRange = args
            except ValueError:
                nrrd = args[0]
                blind8BitRange = False
            self._cnrrdrange = c.nrrdRangeNewSet(nrrd._cnrrd, blind8BitRange)


# NrrdRangeC is for things that are created on the C side of
# the Python-C boundary - we simply keep the reference

class NrrdRangeC(NrrdRangeBase):

    def __init__(self, ctype_obj):
        self._cnrrdrange = ctype_obj
    

##############################################################################
# NrrdKernelSpec class

class NrrdKernelSpecBase(object):

    def __del__(self):
        c.nrrdKernelSpecNix(self._cnrrdkernelspec)

    def _get_ctype(self):
        return self._cnrrdkernelspec
    _ctype = property(_get_ctype)

    ##########################################################################

    def __copy__(self):
        new_cnrrdkernelspec = c.nrrdKernelSpecCopy(self._cnrrdkernelspec)
        return NrrdKernelSpecC(new_cnrrdkernelspec)

    ##########################################################################

    def set(self, kernel, parameters):
        params = (c_double * c.NRRD_KERNEL_PARMS_NUM)()
        for ix, v in enumerate(parameters):
            params[ix] = v
        c.kernelSpecSet(self._cnrrdkernelspec, kernel._cnrrdkernel, params)
        
class NrrdKernelSpec(NrrdKernelSpecBase):

    def __init__(self):
        self._cnrrdkernelspec = c.nrrdKernelSpecNew()

class NrrdKernelSpecC(NrrdKernelSpecBase):

    def __init__(self, cobj):
        self._cnrrdkernelspec = cobj


##############################################################################
# NrrdResampleContext class

class NrrdResampleContext(object):

    def __init__(self):
        self._cnrrdresamplecontext = c.nrrdResampleContextNew()

    def __del__(self):
        c.nrrdResampleContextNix(self._cnrrdresamplecontext)

    def _get_ctype(self):
        return self._cnrrdresamplecontext
    _ctype = property(_get_ctype)

    ##########################################################################

    def set_default_center(self, defaultCenter):
        c.nrrdResampleDefaultCenterSet(self._cnrrdresamplecontext, defaultCenter)

    def set_nrrd(self, nrrd):
        c.nrrdResampleNrrdSet(self._cnrrdresamplecontext, nrrd._cnrrd)

    def set_kernel(self, axis, kernel, kernel_parms):
        v1 = (c_double * c.NRRD_KERNEL_PARMS_NUM)()
        for i, v in enumerate(kernel_parms):
            v1[i] = v
        c.nrrdResampleKernelSet(self._cnrrdresamplecontext, axis, kernel, v1)
        # Is this right? Mimicking unrrdu/resample.c, sort of
        self.set_full_range(axis)

    def set_samples(self, axis, samples_out):
        c.nrrdResampleSamplesSet(self._cnrrdresamplecontext, axis, samples_out)
        # Is this right? Mimicking unrrdu/resample.c, sort of
        self.set_full_range(axis)

    def set_full_range(self, axis):
        c.nrrdResampleRangeFullSet(self._cnrrdresamplecontext, axis)

    # not quite sure what this does
    def set_range(self, axis, min, max):
        c.nrrdResampleRangeSet(self._cnrrdresamplecontext,
                               axis, min, max)

    def set_boundary(self, boundary):
        c.nrrdResampleBoundarySet(self._cnrrdresamplecontext,
                                  boundary)

    def set_type_out(self, type_out):
        c.nrrdResampleTypeOutSet(self._cnrrdresamplecontext,
                                 type_out)

    def set_pad_value(self, pad_value):
        c.nrrdResamplePadValueSet(self._cnrrdresamplecontext,
                                  pad_value)

    def set_renormalize(self, renormalize):
        c.nrrdResampleRenormalizeSet(self._cnrrdresamplecontext,
                                     renormalize)

    def execute(self):
        r = Nrrd(True)
        c.nrrdResampleExecute(self._cnrrdresamplecontext,
                              r._cnrrd)
        return r

##############################################################################
# Kernels, etc

class Kernel(object):

    TMF              = c.nrrdKernelTMF
    TMF_maxD         = c.nrrdKernelTMF_maxD
    TMF_maxC         = c.nrrdKernelTMF_maxC
    TMF_maxA         = c.nrrdKernelTMF_maxA
    Hann             = c.nrrdKernelHann
    HannD            = c.nrrdKernelHannD
    HannDD           = c.nrrdKernelHannDD
    Blackman         = c.nrrdKernelBlackman
    BlackmanD        = c.nrrdKernelBlackmanD
    BlackmanDD       = c.nrrdKernelBlackmanDD
    Zero             = c.nrrdKernelZero
    Box              = c.nrrdKernelBox
    Cheap            = c.nrrdKernelCheap
    HermiteFlag      = c.nrrdKernelHermiteFlag
    Tent             = c.nrrdKernelTent
    ForwDiff         = c.nrrdKernelForwDiff
    CentDiff         = c.nrrdKernelCentDiff
    BCCubic          = c.nrrdKernelBCCubic
    BCCubicD         = c.nrrdKernelBCCubicD
    BCCubicDD        = c.nrrdKernelBCCubicDD
    AQuartic         = c.nrrdKernelAQuartic
    AQuarticD        = c.nrrdKernelAQuarticD
    AQuarticDD       = c.nrrdKernelAQuarticDD
    C3Quintic        = c.nrrdKernelC3Quintic
    C3QuinticD       = c.nrrdKernelC3QuinticD
    C3QuinticDD      = c.nrrdKernelC3QuinticDD
    Gaussian         = c.nrrdKernelGaussian
    GaussianD        = c.nrrdKernelGaussianD
    GaussianDD       = c.nrrdKernelGaussianDD
    DiscreteGaussian = c.nrrdKernelDiscreteGaussian

class Center(object):
    Node = c.nrrdCenterNode
    Cell = c.nrrdCenterCell   

##############################################################################
# NrrdIter

class NrrdIter(object):

    def __init__(self):
        self._cnrrditer = c.nrrdIterNew()

    def __del__(self):
        c.nrrdIterNix(self._cnrrditer)

    def _get_ctype(self):
        return self._cnrrditer
    _ctype = property(_get_ctype)

    ##########################################################################

    def set_value(self, val):
        c.nrrdIterSetValue(self._ctype, val)

    def set_nrrd(self, nrrd):
        c.nrrdIterSetNrrd(self._ctype, nrrd._ctype)

    def set_own_nrrd(self, nrrd):
        print >>sys.stderr, "Warning, this will probably double-free the nrrd"
        c.nrrdIterSetOwnNrrd(self._ctype, nrrd._ctype)

    def value(self):
        return c.nrrdIterValue(self._ctype)

    def content(self):
        return c.nrrdIterContent(self._ctype)

##############################################################################
# Nrrd class

class NrrdBase(object):

    ##########################################################################
    # Constructors

    # this allows us to say
    # >>> x = Nrrd.load('foo.nrrd')
    # >>> y = numpy.array(x)
    #
    # It creates a copy of the nrrd.
    #
    # see http://numpy.scipy.org/array_interface.shtml
    def _get_array_interface(self):
        nrrd = self._cnrrd.contents
        s = []
        for i in reversed(range(nrrd.dim)):
            s.append(nrrd.axis[i].size)
        r = {'shape': tuple(s),
             'typestr': teem.capi.numpy.numpy_type_map[nrrd.type],
             'data': (nrrd.data, False),
             'version': 3}
        return r
    __array_interface__ = property(_get_array_interface)

    @staticmethod
    def from_array(array):
        """from_array(numpy array, copy=False) -> Nrrd. 
        Makes a copy of the array and takes ownership of the
        new data.
        """
        r = Nrrd(True)
        teem.capi.numpy.nrrd_from_array(array, r._cnrrd)
        return r

    @staticmethod
    def load(filename):
        """load(filename) -> Nrrd
        """

        r = Nrrd(True)
        c.nrrdLoad(r._cnrrd, filename, None)
        return r

    def __copy__(self):
        result = c.nrrdCopy(self._cnrrd)
        return NrrdC(result, True)

    ##########################################################################
    # Destructor

    def __del__(self):
        if self._own_data:
            c.nrrdEmpty(self._cnrrd)
        c.nrrdNix(self._cnrrd)

    ##########################################################################
    # ctypes

    def _get_ctype(self):
        return self._cnrrd
    _ctype = property(_get_ctype)

    ##########################################################################
    # IO

    def save(self, filename, iostate=None):
        try:
            iostate = iostate._ctype
        except AttributeError:
            pass
        c.nrrdSave(filename, self._cnrrd, iostate)

    ##########################################################################
    # Accessors

    type = ctype_simple_rw_property('type')
    dim = ctype_simple_rw_property('dim')
    content = ctype_simple_rw_property('content')
    sampleUnits = ctype_simple_rw_property('sampleUnits')
    space = ctype_simple_rw_property('space')
    spaceDim = ctype_simple_rw_property('spaceDim')
    oldMin = ctype_simple_rw_property('oldMin')
    oldMax = ctype_simple_rw_property('oldMax')
    axis = ctype_simple_ro_property('axis')

    ##########################################################################
    # arithmetic

    def _perform_unary_op(operation):
        def do_it(self):
            r = Nrrd(True)
            c.nrrdArithUnaryOp(r._cnrrd, operation, self._cnrrd)
            return r
        return do_it

    def _perform_binary_op(operation):
        def do_it(self, other):
            r = Nrrd(True)
            c.nrrdArithBinaryOp(r._cnrrd, operation, self._cnrrd, other._cnrrd)
            return r
        return do_it

    def _perform_ternary_op(operation):
        def do_it(self, other1, other2):
            r = Nrrd(True)
            c.nrrdArithBinaryOp(r._cnrrd, operation, self._cnrrd,
                                other1._cnnrd, other2._cnnrd)
            return r
        return do_it

    negative = _perform_unary_op(c.nrrdUnaryOpNegative)
    reciprocal = _perform_unary_op(c.nrrdUnaryOpReciprocal)
    sin = _perform_unary_op(c.nrrdUnaryOpSin)
    cos = _perform_unary_op(c.nrrdUnaryOpCos)
    tan = _perform_unary_op(c.nrrdUnaryOpTan)
    asin = _perform_unary_op(c.nrrdUnaryOpAsin)
    acos = _perform_unary_op(c.nrrdUnaryOpAcos)
    atan = _perform_unary_op(c.nrrdUnaryOpAtan)
    exp = _perform_unary_op(c.nrrdUnaryOpExp)
    log = _perform_unary_op(c.nrrdUnaryOpLog)
    log2 = _perform_unary_op(c.nrrdUnaryOpLog2)
    log10 = _perform_unary_op(c.nrrdUnaryOpLog10)
    log1p = _perform_unary_op(c.nrrdUnaryOpLog1p)
    sqrt = _perform_unary_op(c.nrrdUnaryOpSqrt)
    cbrt = _perform_unary_op(c.nrrdUnaryOpCbrt)
    erf = _perform_unary_op(c.nrrdUnaryOpErf)
    ceil = _perform_unary_op(c.nrrdUnaryOpCeil)
    floor = _perform_unary_op(c.nrrdUnaryOpFloor)
    round_up = _perform_unary_op(c.nrrdUnaryOpRoundUp)
    round_down = _perform_unary_op(c.nrrdUnaryOpRoundDown)
    abs = _perform_unary_op(c.nrrdUnaryOpAbs)
    sgn = _perform_unary_op(c.nrrdUnaryOpSgn)
    exists = _perform_unary_op(c.nrrdUnaryOpExists)
    rand = _perform_unary_op(c.nrrdUnaryOpRand)
    normal_rand = _perform_unary_op(c.nrrdUnaryOpNormalRand)
    if_ = _perform_unary_op(c.nrrdUnaryOpIf)
    zero = _perform_unary_op(c.nrrdUnaryOpZero)
    one = _perform_unary_op(c.nrrdUnaryOpOne)
    del _perform_unary_op

    add = _perform_binary_op(c.nrrdBinaryOpAdd)
    subtract = _perform_binary_op(c.nrrdBinaryOpSubtract)
    multiply = _perform_binary_op(c.nrrdBinaryOpMultiply)
    divide = _perform_binary_op(c.nrrdBinaryOpDivide)
    pow = _perform_binary_op(c.nrrdBinaryOpPow)
    sgnPow = _perform_binary_op(c.nrrdBinaryOpSgnPow)
    mod = _perform_binary_op(c.nrrdBinaryOpMod)
    fmod = _perform_binary_op(c.nrrdBinaryOpFmod)
    atan2 = _perform_binary_op(c.nrrdBinaryOpAtan2)
    min = _perform_binary_op(c.nrrdBinaryOpMin)
    max = _perform_binary_op(c.nrrdBinaryOpMax)
    LT = _perform_binary_op(c.nrrdBinaryOpLT)
    LTE = _perform_binary_op(c.nrrdBinaryOpLTE)
    GT = _perform_binary_op(c.nrrdBinaryOpGT)
    GTE = _perform_binary_op(c.nrrdBinaryOpGTE)
    compare = _perform_binary_op(c.nrrdBinaryOpCompare)
    equal = _perform_binary_op(c.nrrdBinaryOpEqual)
    notEqual = _perform_binary_op(c.nrrdBinaryOpNotEqual)
    exists2 = _perform_binary_op(c.nrrdBinaryOpExists)
    if2 = _perform_binary_op(c.nrrdBinaryOpIf)
    del _perform_binary_op

    add3 = _perform_ternary_op(c.nrrdTernaryOpAdd)
    multiply3 = _perform_ternary_op(c.nrrdTernaryOpMultiply)
    min3 = _perform_ternary_op(c.nrrdTernaryOpMin)
    max3 = _perform_ternary_op(c.nrrdTernaryOpMax)
    clamp = _perform_ternary_op(c.nrrdTernaryOpClamp)
    ifElse = _perform_ternary_op(c.nrrdTernaryOpIfElse)
    lerp = _perform_ternary_op(c.nrrdTernaryOpLerp)
    exists3 = _perform_ternary_op(c.nrrdTernaryOpExists)
    inOpen = _perform_ternary_op(c.nrrdTernaryOpInOpen)
    inClosed = _perform_ternary_op(c.nrrdTernaryOpInClosed)
    del _perform_ternary_op

    gamma = nrrd_fun_make_new_nrrd(c.nrrdArithGamma)
    convert = nrrd_fun_make_new_nrrd(c.nrrdConvert)
    clampConvert = nrrd_fun_make_new_nrrd(c.nrrdClampConvert)
    quantize = nrrd_fun_make_new_nrrd(c.nrrdQuantize)
    unquantize = nrrd_fun_make_new_nrrd(c.nrrdUnquantize)

    # nrrdArithIter*
    @staticmethod
    def arith_iter_binary_op(op, in1, in2):
        result = Nrrd(True)
        c.nrrdArithIterBinaryOp(result._ctype,
                                op,
                                in1._ctype,
                                in2._ctype)
        return result

    @staticmethod
    def arith_iter_ternary_op(op, in1, in2, in3):
        result = Nrrd(True)
        c.nrrdArithIterBinaryOp(result._ctype,
                                op,
                                in1._ctype,
                                in2._ctype,
                                in3._ctype)
        return result

    ##########################################################################

    def histoEq(self, bins, smart, amount):
        r = Nrrd(True)
        c.nrrdHistoEq(r._cnrrd, self._cnrrd, None,
                      bins, smart, amount)
        return r

    def histoEq_store_map(self, bins, smart, amount):
        """same as histoEq, but returns a pair of nrrds, where the
        first is the equalized volume, and the second is the
        equalization map."""
        r = Nrrd(True)
        rmap = Nrrd(True)
        c.nrrdHistoEq(r._cnrrd, self._cnrrd, byref(rmap),
                      bins, smart, amount)
        return (r, rmap)

    apply1DLut = nrrd_fun_make_new_nrrd(c.nrrdApply1DLut)
    applyMulti1DLut = nrrd_fun_make_new_nrrd(c.nrrdApplyMulti1DLut)
    apply1DRegMap = nrrd_fun_make_new_nrrd(c.nrrdApply1DRegMap)
    applyMulti1DRegMap = nrrd_fun_make_new_nrrd(c.nrrdApplyMulti1DRegMap)
    apply1DIrregMap = nrrd_fun_make_new_nrrd(c.nrrdApply1DIrregMap)
    apply1DSubstitution = nrrd_fun_make_new_nrrd(c.nrrdApply1DSubstitution)
    apply2DLut = nrrd_fun_make_new_nrrd(c.nrrdApply2DLut)
    slice = nrrd_fun_make_new_nrrd(c.nrrdSlice)

    def crop(self, mins, maxs):
        r = Nrrd(True)
        t = c_size_t * self.dim
        v1 = t()
        v2 = t()
        for ix, v in enumerate(mins):
            v1[ix] = v
        for ix, v in enumerate(maxs):
            v2[ix] = v
        c.nrrdCrop(r._cnrrd, self._cnrrd, v1, v2)
        return r

    simpleCrop = nrrd_fun_make_new_nrrd(c.nrrdSimpleCrop)
    splice = nrrd_fun_make_new_nrrd(c.nrrdSplice)
    
    cheapMedian = nrrd_fun_make_new_nrrd(c.nrrdCheapMedian)

    def pad(self, mins, maxs, boundary, value):
        r = Nrrd(True)
        t = c_size_t * self.dim
        v1 = t()
        v2 = t()
        for ix, v in enumerate(mins):
            v1[ix] = v
        for ix, v in enumerate(maxs):
            v2[ix] = v
        c.nrrdPad(r._cnrrd, self._cnrrd, v1, v2, boundary, value)
        return r

    simplePad = nrrd_fun_make_new_nrrd(c.nrrdSimplePad_nva)

    def inset(self, inset, mins):
        r = Nrrd(True)
        v1 = (c_size_t * self.dim)()
        for ix, v in enumerate(mins):
            v1[ix] = v
        c.nrrdPad(r._cnrrd, self._cnrrd, inset._cnrrd, v1)
        return r

    project = nrrd_fun_make_new_nrrd(c.nrrdProject)

    _histo = nrrd_fun_make_new_nrrd(c.nrrdHisto)
    def histo(self, bins, weight=None, range=None, type=Type.UInt):
        """histo(bins, weight=None, range=None, type=None) -> Nrrd

        returns histogram. Note that calling order is slightly different from
        c version to allow for default parameters.
        """
        return self._histo(range, weight, bins, type)


    _histoDraw = nrrd_fun_make_new_nrrd(c.nrrdHistoDraw)
    def histoDraw(self, height, showLog=1, max=0.0):
        return self._histoDraw(height, showLog, max)

    histoAxis = nrrd_fun_make_new_nrrd(c.nrrdHistoAxis)
    def histoThresholdOtsu(self, exponent):
        r = c_double(0.0)
        c.nrrdHistoThresholdOtsu(by_ref(c_double), self._cnrrd, exponent)
        return r.value

    _project = nrrd_fun_make_new_nrrd(c.nrrdProject)

    def project(self, axis, measure, type=None):
        if type == None:
            type = self.type
        return self._project(axis, measure, type)

    def _perform_projection(measure):
        def do_it(self, axis, type=None):
            return self.project(axis, measure, type)
        return do_it
    project_max   = _perform_projection(c.nrrdMeasureMax)
    mean          = _perform_projection(c.nrrdMeasureMean)
    median        = _perform_projection(c.nrrdMeasureMedian)
    mode          = _perform_projection(c.nrrdMeasureMode)
    product       = _perform_projection(c.nrrdMeasureProduct)
    sum           = _perform_projection(c.nrrdMeasureSum)
    l1            = _perform_projection(c.nrrdMeasureL1)
    l2            = _perform_projection(c.nrrdMeasureL2)
    linf          = _perform_projection(c.nrrdMeasureLinf)
    variance      = _perform_projection(c.nrrdMeasureVariance)
    sd            = _perform_projection(c.nrrdMeasureSD)
    skew          = _perform_projection(c.nrrdMeasureSkew)
    lineSlope     = _perform_projection(c.nrrdMeasureLineSlope)
    lineIntercept = _perform_projection(c.nrrdMeasureLineIntercept)
    lineError     = _perform_projection(c.nrrdMeasureLineError)
    histoMin      = _perform_projection(c.nrrdMeasureHistoMin)
    histoMax      = _perform_projection(c.nrrdMeasureHistoMax)
    histoMean     = _perform_projection(c.nrrdMeasureHistoMean)
    histoMedian   = _perform_projection(c.nrrdMeasureHistoMedian)
    histoMode     = _perform_projection(c.nrrdMeasureHistoMode)
    histoProduct  = _perform_projection(c.nrrdMeasureHistoProduct)
    histoSum      = _perform_projection(c.nrrdMeasureHistoSum)
    histoL2       = _perform_projection(c.nrrdMeasureHistoL2)
    histoVariance = _perform_projection(c.nrrdMeasureHistoVariance)
    histoSD       = _perform_projection(c.nrrdMeasureHistoSD)
    del _perform_projection

    def distanceL2(self, threshold, insideHigher, type=None):
        r = Nrrd(True)
        if type is None:
            type = self.type
        c.nrrdDistanceL2(r._cnrrd, self._cnrrd, type, None, threshold, insideHigher)
        return r

    ##########################################################################
    # numeric types

    @staticmethod
    def _make_iter(v):
        i = NrrdIter()
        if type(v) == Nrrd:
            i.set_nrrd(v)
        else:
            v = float(v)
            i.set_value(v)
        return i
    
    def _make_binary_op_iter(nrrd_method, iter_op):
        def do_it(nrrd1, nrrd2):
            nrrd_iter1 = Nrrd._make_iter(nrrd1)
            nrrd_iter2 = Nrrd._make_iter(nrrd2)
            return Nrrd.arith_iter_binary_op(iter_op,
                                             nrrd_iter1,
                                             nrrd_iter2)
        return do_it

    def _make_binary_op_iter_reverse(nrrd_method, iter_op):
        def do_it(nrrd2, nrrd1):
            nrrd_iter1 = Nrrd._make_iter(nrrd1)
            nrrd_iter2 = Nrrd._make_iter(nrrd2)
            return Nrrd.arith_iter_binary_op(iter_op,
                                             nrrd_iter1,
                                             nrrd_iter2)
        return do_it

    __add__ = _make_binary_op_iter(add, c.nrrdBinaryOpAdd)
    __sub__ = _make_binary_op_iter(subtract, c.nrrdBinaryOpSubtract)
    __mul__ = _make_binary_op_iter(multiply, c.nrrdBinaryOpMultiply)
    __div__ = _make_binary_op_iter(divide, c.nrrdBinaryOpDivide)
    __pow__ = _make_binary_op_iter(pow, c.nrrdBinaryOpPow)
    
    __radd__ = _make_binary_op_iter_reverse(add, c.nrrdBinaryOpAdd)
    __rsub__ = _make_binary_op_iter_reverse(subtract, c.nrrdBinaryOpSubtract)
    __rmul__ = _make_binary_op_iter_reverse(multiply, c.nrrdBinaryOpMultiply)
    __rdiv__ = _make_binary_op_iter_reverse(divide, c.nrrdBinaryOpDivide)
    __rpow__ = _make_binary_op_iter_reverse(pow, c.nrrdBinaryOpPow)


class Nrrd(NrrdBase):
    
    def __init__(self, will_own_data=False):
        self._cnrrd = c.nrrdNew()
        self._own_data = will_own_data

class NrrdC(NrrdBase):

    def __init__(self, c_obj, will_own_data=False):
        self._cnrrd = c_obj
        self._own_data = will_own_data

##############################################################################
# NrrdIoState class

class NrrdIoState(object):
    def __init__(self, encoding=None, format=None):
        self._cnrrdiostate = c.nrrdIoStateNew()
        if format:
            c.nrrdIoStateFormatSet(self._cnrrdiostate,
                                   format)
        if encoding:
            c.nrrdIoStateEncodingSet(self._cnrrdiostate,
                                     encoding)

    def __del__(self):
        c.nrrdIoStateNix(self._cnrrdiostate)

    def _get_ctype(self):
        return self._cnrrdiostate
    _ctype = property(_get_ctype)
